"""Stack of notes"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_chord.ipynb.

# %% auto 0
__all__ = ['Chord', 'PolyChord']

# %% ../nbs/01_chord.ipynb 3
import numpy as np
import pandas as pd
from fastcore.all import *
from IPython.display import Audio
from mingus.core import chords as mingus_chords

from . import Note, Interval

# %% ../nbs/01_chord.ipynb 5
class Chord(BasicRepr):
    def __init__(self, notes: List[Note]):
        self.notes = [Note(n) if isinstance(n, str) else n for n in notes]
        self.s_notes = [str(n) for n in self.notes]
        self.short_s_notes = [f"{n.note}{n.oct}" for n in self.notes]
        self.first = self.notes[0]
        self.s_first = str(self.first)

    def names(self) -> list[str]: return mingus_chords.determine(self.s_notes)
    def name(self): return self.names()[0] if self.names() else "No chord found."

    @classmethod
    def from_short(cls, c: str): return cls(mingus_chords.from_shorthand(c)) 

    def __repr__(self): return f"Chord: '{self.name()}'. Notes: {self.short_s_notes}"
    def __add__(self, other): return Chord([n + other for n in self.notes])
    def __sub__(self, other): return Chord([n - other for n in self.notes])
    def __mod__(self, other): return Chord([n % other for n in self.notes])
    def __len__(self): return len(self.notes)
    def __floordiv__(self, other): return Chord([n // other for n in self.notes])
    def __iter__(self) -> list[str]: return iter(self.notes)
    def __index__(self):
        mask = 0
        for n in self.notes: mask |= 1 << n.midi
        return mask
    def __int__(self): return self.__index__()
    @property
    def midi(self): return [n.midi for n in self.notes]
    
    def _compare_notes(self, other, op): return all(op(n1, n2) for n1, n2 in zip(self.notes, other.notes))
    def __eq__(self, other): return self.first == other.first and self._compare_notes(other, lambda x, y: x == y)
    def __ne__(self, other): return not self == other
    def __lt__(self, other): return self.first < other.first or (self.first == other.first and self._compare_notes(other, lambda x, y: x < y))
    def __le__(self, other): return self.first < other.first or (self.first == other.first and self._compare_notes(other, lambda x, y: x <= y))
    def __gt__(self, other): return self.first > other.first or (self.first == other.first and self._compare_notes(other, lambda x, y: x > y))
    def __ge__(self, other): return self.first > other.first or (self.first == other.first and self._compare_notes(other, lambda x, y: x >= y))

# %% ../nbs/01_chord.ipynb 27
@patch
def __mul__(self:Note, other: Note):
    """ Multiply two notes to form a chord. """
    return Chord([self, other])

@patch
def __mul__(self:Chord, other):
    if isinstance(other, Note):
        return Chord(self.notes + [other])
    else:
        raise ValueError("Chord objects can only be multiplied with Note or other Chord objects")

# %% ../nbs/01_chord.ipynb 30
@patch
def invert(self:Chord, n: int = 1):
    assert n > 0 and n < len(self.s_notes), f"Invalid inversion '{n}' for chord with '{len(self.s_notes)}' notes."
    return Chord(self.notes[n:] + [Note(str(note), oct=note.oct + 1) for note in self.notes[:n]])

# %% ../nbs/01_chord.ipynb 33
@patch
def rel_intervals(self:Chord):
    return [Interval(self.notes[0], n) for n in self.notes[1:]]

@patch
def abs_intervals(self:Chord):
    return [Interval(n1, n2) for n1, n2 in zip(self.notes, self.notes[1:])]

# %% ../nbs/01_chord.ipynb 37
@patch
def get_audio_array(self:Chord, length=1):
    return np.sum([n.get_audio_array(length) for n in self.notes], axis=0)

@patch
def play(self:Chord, length=1): 
    return Audio(self.get_audio_array(length), rate=44100)

# %% ../nbs/01_chord.ipynb 43
@patch
def to_frame(self:Chord):
    rel_intervals = self.rel_intervals()
    rel_short_intvals = [i.short for i in rel_intervals]
    rel_long_intvals = [i.long for i in rel_intervals]
    abs_intervals = self.abs_intervals()
    abs_short_intvals = [i.short for i in abs_intervals]
    abs_long_intvals = [i.long for i in abs_intervals]
    
    d = {
        "Notes": self.notes,
        "Relative Degree": [1] + rel_short_intvals,
        "Relative Interval": ["unison"] + rel_long_intvals,
        "Absolute Interval": ["unison"] + abs_long_intvals,
        "Absolute Degree": [1] + abs_short_intvals,
    }
    return pd.DataFrame(d)

# %% ../nbs/01_chord.ipynb 48
class PolyChord(Chord):
    def __init__(self, chords: list[Chord]):
        self.chords = chords
        super().__init__([note for chord in chords for note in chord.notes])
    def __repr__(self): return f"PolyChord: '{'|'.join([c.name() for c in self.chords])}'. Notes: {self.short_s_notes}"

# %% ../nbs/01_chord.ipynb 52
@patch
def invert(self:PolyChord, n: int = 1):
    return PolyChord([c.invert(n) for c in self.chords])

# %% ../nbs/01_chord.ipynb 57
@patch
def to_frame(self:PolyChord) -> list[pd.DataFrame]:
    return [c.to_frame() for c in self.chords]



"""Stack of intervals"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_scale.ipynb.

# %% auto 0
__all__ = ['INTERVAL_NAMES', 'INTERVAL_HALF_STEPS', 'SCALES', 'd', 'INV_SCALES_BY_INTERVAL', 'INV_SCALES_BY_NAME', 'Scale',
           'semi_to_intvals']

# %% ../nbs/02_scale.ipynb 3
import numpy as np
import pandas as pd
from fastcore.all import *
from IPython.display import Audio
from itertools import combinations, accumulate
from collections import defaultdict

from . import Note, Chord, Interval

# %% ../nbs/02_scale.ipynb 4
INTERVAL_NAMES = {
    "1": "unison",
    "b2": "minor second",
    "2": "major second",
    "#2": "augmented second",
    "b3": "minor third",
    "3": "major third",
    "#3": "augmented third",
    "b4": "minor fourth",
    "4": "perfect fourth",
    "#4": "augmented fourth",
    "b5": "diminished fifth",
    "5": "perfect fifth",
    "#5": "augmented fifth",
    "b6": "minor sixth",
    "6": "major sixth",
    "#6": "augmented sixth",
    "b7": "minor seventh",
    "7": "major seventh",
    "8": "octave",
    "b9": "minor ninth",
    "9": "major ninth",
    "#9": "augmented ninth",
    "b10": "minor tenth",
    "10": "major tenth",
    "b11": "minor eleventh",
    "11": "major eleventh",
    "#11": "augmented eleventh",
    "b12": "minor twelfth",
    "12": "major twelfth",
    "b13": "minor thirteenth",
    "13": "major thirteenth",
    "#13": "augmented thirteenth"
}
INTERVAL_HALF_STEPS = {
    "1": 0,
    "#1": 1,
    "b2": 1,
    "2": 2,
    "#2": 3,
    "b3": 3,
    "3": 4,
    "b4": 4,
    "4": 5,
    "#4": 6,
    "b5": 6,
    "5": 7,
    "#5": 8,
    "b6": 8,
    "6": 9,
    "#6": 10,
    "b7": 10,
    "7": 11,
    "8": 12,
    "b9": 13,
    "9": 14,
    "#9": 15,
    "10": 16,
    "11": 17,
    "#11": 18,
    "12": 19,
    "b13": 20,
    "13": 21,
    "#13": 22,
    "14": 23,
    "15": 24
}


SCALES = {# Major modes
          "ionian": ["1", "2", "3", "4", "5", "6", "7"],
          "major": ["1", "2", "3", "4", "5", "6", "7"],
          "minor": ["1", "2", "b3", "4", "5", "b6", "b7"],
          "natural minor": ["1", "2", "b3", "4", "5", "b6", "b7"],
          "dorian": ["1", "2", "b3", "4", "5", "6", "b7"],
          "phrygian": ["1", "b2", "b3", "4", "5", "b6", "b7"],
          "lydian": ["1", "2", "3", "#4", "5", "6", "7"],
          "mixolydian": ["1", "2", "3", "4", "5", "6", "b7"],
          "aeolian": ["1", "2", "b3", "4", "5", "b6", "b7"],
          "locrian": ["1", "b2", "b3", "4", "b5", "b6", "b7"],
          # Melodic Minor modes
          "melodic minor": ["1", "2", "b3", "4", "5", "6", "7"],
          "dorian b2": ["1", "b2", "b3", "4", "5", "6", "b7"],
          "lydian augmented": ["1", "2", "3", "#4", "#5", "6", "7"],
          "lydian b7": ["1", "2", "3", "#4", "5", "6", "b7"],
          "lydian dominant": ["1", "2", "3", "#4", "5", "6", "b7"],
          "acoustic": ["1", "2", "3", "#4", "5", "6", "b7"],
          "mixolydian #11": ["1", "2", "3", "#4", "5", "6", "b7"],
          "locrian n2": ["1", "2", "b3", "4", "b5", "b6", "7"],
          "altered": ["1", "b2", "b3", "b4", "b5", "b6", "b7"],
          "altered dominant": ["1", "b2", "b3", "b4", "b5", "b6", "b7"],
          "diminished whole-tone": ["1", "b2", "b3", "b4", "b5", "b6", "b7"],
          "dominant whole-tone": ["1", "b2", "b3", "b4", "b5", "b6", "b7"],
          "aeolian dominant": ["1", "2", "3", "4", "5", "b6", "b7"],
          # Harmonic minor modes
          "harmonic minor": ["1", "2", "b3", "4", "5", "b6", "7"],
          "locrian 6": ["1", "b2", "b3", "4", "b5", "6", "b7"],
          "ionian augmented": ["1", "2", "3", "4", "#5", "6", "7"],
          "dorian #4": ["1", "2", "b3", "#4", "5", "6", "b7"],
          "phrygian major": ["1", "b2", "3", "4", "5", "b6", "b7"],
          "phrygian dominant": ["1", "b2", "3", "4", "5", "b6", "b7"],
          "spanish phrygian": ["1", "b2", "3", "4", "5", "b6", "b7"],
          "spanish major": ["1", "b2", "3", "4", "5", "b6", "b7"],
          "spanish": ["1", "b2", "3", "4", "5", "b6", "b7"],
          "lydian #2": ["1", "#2", "3", "#4", "5", "6", "7"],
          "lydian #9": ["1", "#2", "3", "#4", "5", "6", "7"],
          "altered dominant bb7": ["1", "b2", "b3", "b4", "b5", "b6", "6"],
          "altered 13": ["1", "b2", "b3", "b4", "b5", "b6", "6"],
          # Harmonic major modes
          "harmonic major": ["1", "2", "3", "4", "5", "b6", "7"],
          "ionian b6": ["1", "2", "3", "4", "5", "b6", "7"],
          "dorian b5": ["1", "2", "b3", "4", "b5", "6", "b7"],
          "phrygian b4": ["1", "b2", "b3", "b4", "5", "b6", "b7"],
          "lydian b3": ["1", "2", "b3", "#4", "5", "6", "7"],
          "melodic minor #4": ["1", "2", "b3", "#4", "5", "6", "7"],
          "mixolydian b2": ["1", "b2", "3", "4", "5", "6", "b7"],
          # Lydian augmented already defined in melodic minor modes
          "locrian bb7": ["1", "b2", "b3", "4", "b5", "b6", "6"],
          # Double harmonic major modes
          "double harmonic major": ["1", "b2", "3", "4", "5", "b6", "7"],
          "double harmonic": ["1", "b2", "3", "4", "5", "b6", "7"],
          "lydian #2#6": ["1", "#2", "3", "#4", "5", "#6", "7"],
          "ultraphrygian": ["1", "b2", "b3", "b4", "5", "b6", "6"],
          "hungarian minor": ["1", "2", "b3", "#4", "5", "b6", "7"],
          "oriental": ["1", "b2", "3", "4", "b5", "6", "b7"], # Mixolydian b2 with lowered 5
          "ionian aug#2": ["1", "#2", "3", "4", "#5", "6", "7"],
          "locrian bb3bb7": ["1", "b2", "2", "4", "b5", "b6", "6"],
          # Bebop scales
          "bebop dominant": ["1", "2", "3", "4", "5", "6", "#6", "7"],
          "bebop major": ["1", "2", "3", "4", "5", "b6", "6", "7"],
          "bebop blues": ["1", "2", "b3", "3", "4", "5", "6", "b7"],
          "bebop melodic minor": ["1", "2", "b3", "4", "5", "b6", "6", "7"],
          # Pentatonic scales
          "pentatonic major": ["1", "2", "3", "5", "6"],
          "major pentatonic": ["1", "2", "3", "5", "6"],
          "pentatonic minor": ["1", "b3", "4", "5", "b7"],
          "minor pentatonic": ["1", "b3", "4", "5", "b7"],
          "pentatonic blues": ["1", "b3", "4", "b5", "5", "b7"],
          "blues pentatonic": ["1", "b3", "4", "b5", "5", "b7"],
          "pentatonic neutral": ["1", "2", "4", "5", "b7"],
          "neutral pentatonic": ["1", "2", "4", "5", "b7"],
          "pentatonic rock": ["1", "b3", "4", "#5", "b7"],
          "rock pentatonic": ["1", "b3", "4", "#5", "b7"],
          "jue": ["1", "b3", "4", "#5", "b7"], # Same intervals as rock pentatonic
          "pentatonic scottish": ["1", "2", "4", "5", "6"],
          "scottish pentatonic": ["1", "2", "4", "5", "6"],
          # Blues scales
          "blues major": ["1", "2", "b3", "3", "5", "6"],
          "blues minor": ["1", "b3", "4", "b5", "5", "b7"],
          "blues": ["1", "b3", "4", "b5", "5", "b7"], # Shorthand for Blues Minor
          "blues diminished": ["1", "b2", "b3", "3", "b5", "5", "6", "b7"],
          "aux diminished blues": ["1", "b2", "b3", "3", "b5", "5", "6", "b7"], 
          "half-whole": ["1", "b2", "b3", "3", "b5", "5", "6", "b7"],
          "balinese": ["1", "b2", "b3", "3", "b5", "5", "6", "b7"], 
          # Gypsy scales
          "gypsy major": ["1", "b2", "3", "4", "5", "b6", "7"],
          "gypsy minor": ["1", "2", "b3", "b5", "5", "b6", "7"],
          "spanish gypsy": ["1", "b2", "3", "4", "5", "b6", "b7"], # Same as Phrygian dominant
          "hungarian gypsy": ["1", "2", "b3", "#4", "5", "b6", "b7"], # Hungarian minor with lowered 7
          # Diminished scales
          "diminished": ["1", "2", "b3", "4", "b5", "b6", "6", "7"],
          "tonic diminished": ["1", "2", "b3", "4", "b5", "b6", "6", "7"],
          "whole-half": ["1", "2", "b3", "4", "#4", "#5", "6", "7"],
          "dominant diminished": ["1", "b2", "b3", "3", "#4", "5", "6", "b7"],
          "whole": ["1", "2", "3", "#4", "#5", "#6"],
          "whole-tone": ["1", "2", "3", "#4", "#5", "#6"],
          "aux augmented": ["1", "2", "3", "#4", "#5", "#6"],
          # Maqam
          "bayati shuri": ["1", "2", "b3", "4", "b5", "6", "b7"],
          "hijaz": ["1", "b2", "3", "4", "5", "b6", "b7"],
          "hijaz kar": ["1", "b2", "3", "4", "5", "6", "7"],
          "huzam": ["1", "b2", "b3", "3", "5", "b6", "b7"],
          "nikriz": ["1", "2", "3", "#4", "5", "6", "7"],
          "tunisian": ["1", "2", "3", "#4", "5", "6", "7"],
          "saba": ["1", "2", "b3", "3", "5", "b6", "b7"],
          "sabah": ["1", "2", "b3", "3", "5", "b6", "b7"],
          "suznak": ["1", "2", "3", "4", "5", "#5", "7"],
          "neveseri": ["1", "b2", "b3", "3", "5", "b6", "b7", "7"],
          # Greek
          "pireaus": ["1", "b2", "3", "#4", "5", "#5", "7"],
          "tsinganikos": ["1", "b2", "3", "4", "#4", "6", "#6"],
          # Hindustani
          "marwa": ["1", "b2", "3", "#4", "5", "6", "7"],
          "poorvi": ["1", "b2", "3", "#4", "5", "6", "7"],
          "segah": ["1", "#2", "3", "4", "5", "#5", "7"],
          "todi": ["1", "b2", "b3", "#4", "5", "b6", "7"],
          # Carnatic
          "charukeshi": ["1", "2", "3", "4", "5", "b6", "b7"],
          "dharmaavati": ["1", "2", "b3", "b5", "5", "6", "7"],
          "lataangi": ["1", "2", "3", "#4", "5", "6", "7"],
          "vachaspati": ["1", "2", "3", "#4", "5", "6", "#6"],
          "natakpriya": ["1", "b2", "b3", "4", "5", "6", "b7"],
          "rampriya": ["1", "b2", "3", "#4", "5", "6", "#6"],
          "suryakant": ["1", "b2", "3", "4", "5", "6", "7"],
          # Japanese
          "joshi akikaze": ["1", "2", "b3", "5", "6"],
          "akikaze joshi": ["1", "2", "b3", "5", "6"],
          "joshi hira": ["1", "2", "b3", "5", "b6"],
          "hira joshi": ["1", "2", "b3", "5", "b6"],
          "joshi iwato": ["1", "b2", "4", "b5", "b7"],
          "iwato joshi": ["1", "b2", "4", "b5", "b7"],
          "joshi kokin": ["1", "2", "4", "5", "b6"],
          "kokin joshi": ["1", "2", "4", "5", "b6"],
          "joshi kumoi": ["1", "b2", "4", "5", "b6"],
          "kumoi joshi": ["1", "b2", "4", "5", "b6"],
          "joshi okinawa": ["1", "3", "4", "5", "7"],
          "okinawa joshi": ["1", "3", "4", "5", "7"],
          "sen in": ["1", "b2", "4", "5", "b7"],
          "in sen": ["1", "b2", "4", "5", "b7"],
          # Misc.
          "chromatic": ["1", "b2", "2", "b3", "3", "4", "#4", "5", "#5", "6", "b7", "7"],
          "augmented": ["1", "#2", "3", "5", "#5", "7"],
          "enigmatic ascending": ["1", "b2", "3", "#4", "#5", "#6", "7"],
          "enigmatic descending": ["1", "b2", "3", "4", "b6", "b7", "7"],
          "hungarian major": ["1", "b3", "3", "b5", "5", "6", "b7"],
          "neapolitan major": ["1", "b2", "b3", "4", "5", "6", "7"],
          "neapolitan minor": ["1", "b2", "b3", "4", "5", "b6", "7"],
          "prometheus": ["1", "2", "3", "b5", "6", "b7"],
          "mystic": ["1", "2", "3", "b5", "6", "b7"],
          "prometheus neapolitan": ["1", "b2", "3", "b5", "6", "b7"], # Prometheus with lowered 2
          "spanish 8 tone": ["1", "b2", "b3", "3", "4", "b5", "b6", "b7"],
          }

# Invert scales so alternative names can be identified by intervals or name
d = defaultdict(list)
[d[tuple(v)].append(k) for k,v in SCALES.items()]
INV_SCALES_BY_INTERVAL = dict(d)
INV_SCALES_BY_NAME = {k: d[tuple(v)] for k,v in SCALES.items()}

# %% ../nbs/02_scale.ipynb 7
class Scale:
    def __init__(self, name: str):
        self.name = name.lower()
        self.intervals = SCALES.get(self.name, [])
        self.flats = sum(1 for i in self.intervals if i.startswith("b"))
        self.sharps = sum(1 for i in self.intervals if i.startswith("#"))
        self.naturals = len(self.intervals) - self.flats - self.sharps

    @classmethod
    def available_scales(cls): return list(SCALES.keys())
    
    @classmethod
    def from_intervals(cls, intervals: list[str], name: str = None):
        """Create a custom scale from a list of intervals."""
        for i in intervals:
            assert i in list(INTERVAL_NAMES), f"Interval '{i}' not valid. Available intervals: '{list(INTERVAL_NAMES.keys())}'"
        # Infer name if not given
        custom_scale = cls(INV_SCALES_BY_INTERVAL.get(tuple(intervals), ["unknown"])[0] if not name else name)
        custom_scale.intervals = intervals
        return custom_scale
    
    @property
    def rel_semitones(self):
        return [INTERVAL_HALF_STEPS[interval] for interval in self.intervals]
    
    @property
    def abs_semitones(self):
        rel = self.rel_semitones
        abs = []
        for i, r in enumerate(rel[1:]):
            abs.append(r - rel[i])
        # Last remaining interval
        abs.append(12-sum(abs))
        return abs
    
    def __repr__(self): return f"Scale: {self.name.title()}. Intervals: {self.intervals}"
    def __eq__(self, other): return self.intervals == other.intervals
    def __ne__(self, other): return not self == other
    def __iter__(self) -> list[str]: return iter(self.intervals)
    def __len__(self): return len(self.intervals)

# %% ../nbs/02_scale.ipynb 31
@patch
def get_notes(self:Scale, root, oct=4):
    """Get the notes of a scale from a root note."""
    root = Note(root, oct=oct) if isinstance(root, str) else root
    return [root + int(INTERVAL_HALF_STEPS[i]) for i in self.intervals]

# %% ../nbs/02_scale.ipynb 36
@patch
def get_diatonic_chords(self:Scale, root, min_notes=3):
    assert min_notes > 1, "min_notes must be greater than 1."
    notes = self.get_notes(root)
    return [Chord(combo) for n in range(min_notes, len(notes)+1) for combo in combinations(notes, n)]

# %% ../nbs/02_scale.ipynb 40
@patch
def rel_interval_names(self:Scale, short=False):
        return self.intervals if short else [INTERVAL_NAMES[i] for i in self.intervals[1:]]

@patch
def abs_interval_names(self:Scale, short=False):
    return [getattr(Interval.from_semitones(i), "long" if not short else "short") for i in self.abs_semitones]

# %% ../nbs/02_scale.ipynb 44
@patch
def get_scale_names(self:Scale):
    return INV_SCALES_BY_INTERVAL.get(tuple(self.intervals), [])

# %% ../nbs/02_scale.ipynb 47
@patch
def _shift_abs_semitones(self:Scale, n):
    """ Shift the absolute semitones of a scale by n steps. """
    return self.abs_semitones[n:] + self.abs_semitones[:n]

@patch
def _shift_rel_semitones(self:Scale, n:int):
    """ Shift relative semitones by n steps. """
    return [0] + list(accumulate(self._shift_abs_semitones(n)[:-1]))

# %% ../nbs/02_scale.ipynb 52
def semi_to_intvals(semitones):
    """Convert relative semitone values to interval names."""
    # Major scale is the baseline for accidentals
    maj = [0, 2, 4, 5, 7, 9, 11, 12, 14, 16, 17, 19, 21, 23, 24, 26, 28, 29, 31, 33, 35]
    acc = {-1: 'b', 0: '', 1: '#'}
    intervals = []
    for i, s in enumerate(semitones):
        deg = i + 1
        diff = (s - maj[i]) % 12
        if diff > 6: diff -= 12
        if diff == 2: intervals.append(f'{deg+1}')
        elif diff == -2: intervals.append(f'{deg-1}')
        else: intervals.append(f'{acc.get(diff, "")}{deg}')
    return intervals

# %% ../nbs/02_scale.ipynb 57
@patch
def shift_intvals(self:Scale, n:int) -> list[str]:
    """Shift the intervals of a scale by n steps."""
    return semi_to_intvals(self._shift_rel_semitones(n))

@patch
def get_modes(self:Scale) -> list[Scale]:
    return [Scale.from_intervals(self.shift_intvals(i)) for i in range_of(self)]

# %% ../nbs/02_scale.ipynb 67
@patch
def get_audio_array(self:Scale, root, oct=4, length=0.3):
    notes = self.get_notes(root, oct=oct)
    octave = Note(root, oct=oct+1).get_audio_array(length=length)
    return np.concatenate([n.get_audio_array(length) for n in notes] + [octave])

@patch
def play(self:Scale, root, oct=4, length=0.3): 
    return Audio(self.get_audio_array(root, oct=oct, length=length), rate=44100)

# %% ../nbs/02_scale.ipynb 76
@patch
def get_triads(self:Scale, root):
    """Get all triads in scale starting from root note."""
    notes = self.get_notes(root)
    return [Chord([notes[i], 
                  Note(str(notes[(i+2)%7]), oct=notes[i].oct + (i+2)//7),
                  Note(str(notes[(i+4)%7]), oct=notes[i].oct + (i+4)//7)]) 
            for i in range(len(notes))]

# %% ../nbs/02_scale.ipynb 80
@patch
def play_triads(self:Scale, root):
    """Play all triads in scale starting from root note."""
    return Audio(np.concatenate([c.get_audio_array() for c in self.get_triads(root)]), rate=44100)

# %% ../nbs/02_scale.ipynb 83
@patch
def get_sevenths(self:Scale, root):
    """Get all seventh chords in scale starting from root note."""
    notes = self.get_notes(root)
    return [Chord([notes[i], 
                  Note(str(notes[(i+2)%7]), oct=notes[i].oct + (i+2)//7),
                  Note(str(notes[(i+4)%7]), oct=notes[i].oct + (i+4)//7),
                  Note(str(notes[(i+6)%7]), oct=notes[i].oct + (i+6)//7)]) 
            for i in range(len(notes))]

# %% ../nbs/02_scale.ipynb 86
@patch
def play_sevenths(self:Scale, root):
    """Play all seventh chords in scale starting from root note."""
    return Audio(np.concatenate([c.get_audio_array() for c in self.get_sevenths(root)]), rate=44100)

# %% ../nbs/02_scale.ipynb 90
@patch
def to_frame(self:Scale, root=None):
    d = {
        "Degree": self.intervals,
        "Relative Interval": ["unison"] + self.rel_interval_names(),
        "Mode": [m.name for m in self.get_modes()],
        "Relative Semitones": self.rel_semitones,
        "Absolute Semitones": self.abs_semitones,
    }
    if root:
        d.update({
            "Notes": self.get_notes(root),
            "Triad": [t.name() for t in self.get_triads(root)],
            "Seventh Chord": [s.name() for s in self.get_sevenths(root)],
        })
    return pd.DataFrame(d)



"""Visualization tools such as piano keys and guitar fretboards."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_viz.ipynb.

# %% auto 0
__all__ = ['Piano', 'Guitar']

# %% ../nbs/03_viz.ipynb 3
from fasthtml import *
from fastcore.all import *
from fasthtml.jupyter import render_ft
from IPython.display import HTML

from . import Note, Chord, Scale

render_ft()

# %% ../nbs/03_viz.ipynb 6
class Piano:
    def __ft__(self, highlight: list[Note] = None):
        notes = [str(n) for n in listify(highlight)] if highlight else []
        css = """
    <style>
    .piano { background: #222; padding: 20px 0; position: relative; width: 480px; }
    .white-keys { display: flex; }
    .white-key {
        width: 40px; height: 125px; background: #fff;
        border: 1px solid #000;
        color: #111; font-size: 18px; text-align: center; line-height: 200px; font-family: Arial;
        position: relative; z-index: 1;
    }
    .black-key {
        width: 20px; height: 80px; background: #000; color: #fff;
        border: 1px solid #333; position: absolute; z-index: 2;
        text-align: center; line-height: 100px; font-family: Arial; font-size: 14px;
        left: 0; top: 20px; pointer-events: none;
    }
    .highlight { background: #ff0 !important; color: #000 !important; }
    .highlight-black { background: #ff0 !important; color: #000 !important; }
    </style>
    """
        white_notes = ['C', 'D', 'E', 'F', 'G', 'A', 'B', 'C', 'D', 'E', 'F']
        black_keys = [
            (0, ('C#', 'Db')), (1, ('D#', 'Eb')), (3, ('F#', 'Gb')), (4, ('G#', 'Ab')), (5, ('A#', 'Bb')),
            (7, ('C#', 'Db')), (8, ('D#', 'Eb')), (10, ('F#', 'Gb'))
        ]
        html = f'<div class="piano" style="width:{len(white_notes)*40}px">'
        html += '<div class="white-keys">'
        for note in white_notes:
            cls = "white-key"
            if note in notes:
                cls += " highlight"
            html += f'<div class="{cls}">{note}</div>'
        for idx, (sharp, flat) in black_keys:
            cls = "black-key"
            if sharp in notes or flat in notes:
                cls += " highlight-black"
            left = (idx + 1) * 40 - 14
            html += f'<div class="{cls}" style="left:{left}px">{sharp}</div>'
        html += '</div>'
        return HTML(css + html)
    
    def __call__(self, highlight=list[Note]):
        return self.__ft__(highlight)

# %% ../nbs/03_viz.ipynb 20
class Guitar:
    def __ft__(self, highlight: list[Note] = None):
        ...



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/musy',
                'doc_host': 'https://CarloLepelaars.github.io',
                'git_url': 'https://github.com/CarloLepelaars/musy',
                'lib_path': 'musy'},
  'syms': { 'musy.chord': { 'musy.chord.Chord': ('chord.html#chord', 'musy/chord.py'),
                            'musy.chord.Chord.__add__': ('chord.html#chord.__add__', 'musy/chord.py'),
                            'musy.chord.Chord.__eq__': ('chord.html#chord.__eq__', 'musy/chord.py'),
                            'musy.chord.Chord.__floordiv__': ('chord.html#chord.__floordiv__', 'musy/chord.py'),
                            'musy.chord.Chord.__ge__': ('chord.html#chord.__ge__', 'musy/chord.py'),
                            'musy.chord.Chord.__gt__': ('chord.html#chord.__gt__', 'musy/chord.py'),
                            'musy.chord.Chord.__index__': ('chord.html#chord.__index__', 'musy/chord.py'),
                            'musy.chord.Chord.__init__': ('chord.html#chord.__init__', 'musy/chord.py'),
                            'musy.chord.Chord.__int__': ('chord.html#chord.__int__', 'musy/chord.py'),
                            'musy.chord.Chord.__iter__': ('chord.html#chord.__iter__', 'musy/chord.py'),
                            'musy.chord.Chord.__le__': ('chord.html#chord.__le__', 'musy/chord.py'),
                            'musy.chord.Chord.__len__': ('chord.html#chord.__len__', 'musy/chord.py'),
                            'musy.chord.Chord.__lt__': ('chord.html#chord.__lt__', 'musy/chord.py'),
                            'musy.chord.Chord.__mod__': ('chord.html#chord.__mod__', 'musy/chord.py'),
                            'musy.chord.Chord.__mul__': ('chord.html#chord.__mul__', 'musy/chord.py'),
                            'musy.chord.Chord.__ne__': ('chord.html#chord.__ne__', 'musy/chord.py'),
                            'musy.chord.Chord.__repr__': ('chord.html#chord.__repr__', 'musy/chord.py'),
                            'musy.chord.Chord.__sub__': ('chord.html#chord.__sub__', 'musy/chord.py'),
                            'musy.chord.Chord._compare_notes': ('chord.html#chord._compare_notes', 'musy/chord.py'),
                            'musy.chord.Chord.abs_intervals': ('chord.html#chord.abs_intervals', 'musy/chord.py'),
                            'musy.chord.Chord.from_short': ('chord.html#chord.from_short', 'musy/chord.py'),
                            'musy.chord.Chord.get_audio_array': ('chord.html#chord.get_audio_array', 'musy/chord.py'),
                            'musy.chord.Chord.invert': ('chord.html#chord.invert', 'musy/chord.py'),
                            'musy.chord.Chord.midi': ('chord.html#chord.midi', 'musy/chord.py'),
                            'musy.chord.Chord.name': ('chord.html#chord.name', 'musy/chord.py'),
                            'musy.chord.Chord.names': ('chord.html#chord.names', 'musy/chord.py'),
                            'musy.chord.Chord.play': ('chord.html#chord.play', 'musy/chord.py'),
                            'musy.chord.Chord.rel_intervals': ('chord.html#chord.rel_intervals', 'musy/chord.py'),
                            'musy.chord.Chord.to_frame': ('chord.html#chord.to_frame', 'musy/chord.py'),
                            'musy.chord.Note.__mul__': ('chord.html#note.__mul__', 'musy/chord.py'),
                            'musy.chord.PolyChord': ('chord.html#polychord', 'musy/chord.py'),
                            'musy.chord.PolyChord.__init__': ('chord.html#polychord.__init__', 'musy/chord.py'),
                            'musy.chord.PolyChord.__repr__': ('chord.html#polychord.__repr__', 'musy/chord.py'),
                            'musy.chord.PolyChord.invert': ('chord.html#polychord.invert', 'musy/chord.py'),
                            'musy.chord.PolyChord.to_frame': ('chord.html#polychord.to_frame', 'musy/chord.py')},
            'musy.midi': {'musy.midi.test_midi': ('midi.html#test_midi', 'musy/midi.py')},
            'musy.note': { 'musy.note.Interval': ('note.html#interval', 'musy/note.py'),
                           'musy.note.Interval.__abs__': ('note.html#interval.__abs__', 'musy/note.py'),
                           'musy.note.Interval.__add__': ('note.html#interval.__add__', 'musy/note.py'),
                           'musy.note.Interval.__eq__': ('note.html#interval.__eq__', 'musy/note.py'),
                           'musy.note.Interval.__ge__': ('note.html#interval.__ge__', 'musy/note.py'),
                           'musy.note.Interval.__gt__': ('note.html#interval.__gt__', 'musy/note.py'),
                           'musy.note.Interval.__index__': ('note.html#interval.__index__', 'musy/note.py'),
                           'musy.note.Interval.__init__': ('note.html#interval.__init__', 'musy/note.py'),
                           'musy.note.Interval.__int__': ('note.html#interval.__int__', 'musy/note.py'),
                           'musy.note.Interval.__le__': ('note.html#interval.__le__', 'musy/note.py'),
                           'musy.note.Interval.__lt__': ('note.html#interval.__lt__', 'musy/note.py'),
                           'musy.note.Interval.__ne__': ('note.html#interval.__ne__', 'musy/note.py'),
                           'musy.note.Interval.__repr__': ('note.html#interval.__repr__', 'musy/note.py'),
                           'musy.note.Interval.__sub__': ('note.html#interval.__sub__', 'musy/note.py'),
                           'musy.note.Interval.from_semitones': ('note.html#interval.from_semitones', 'musy/note.py'),
                           'musy.note.Interval.is_consonant': ('note.html#interval.is_consonant', 'musy/note.py'),
                           'musy.note.Interval.is_contextual': ('note.html#interval.is_contextual', 'musy/note.py'),
                           'musy.note.Interval.is_dissonant': ('note.html#interval.is_dissonant', 'musy/note.py'),
                           'musy.note.Interval.is_perfect': ('note.html#interval.is_perfect', 'musy/note.py'),
                           'musy.note.Interval.long': ('note.html#interval.long', 'musy/note.py'),
                           'musy.note.Interval.midi': ('note.html#interval.midi', 'musy/note.py'),
                           'musy.note.Interval.set_semitones': ('note.html#interval.set_semitones', 'musy/note.py'),
                           'musy.note.Interval.short': ('note.html#interval.short', 'musy/note.py'),
                           'musy.note.Interval.type': ('note.html#interval.type', 'musy/note.py'),
                           'musy.note.Note': ('note.html#note', 'musy/note.py'),
                           'musy.note.Note.__add__': ('note.html#note.__add__', 'musy/note.py'),
                           'musy.note.Note.__and__': ('note.html#note.__and__', 'musy/note.py'),
                           'musy.note.Note.__eq__': ('note.html#note.__eq__', 'musy/note.py'),
                           'musy.note.Note.__floordiv__': ('note.html#note.__floordiv__', 'musy/note.py'),
                           'musy.note.Note.__ge__': ('note.html#note.__ge__', 'musy/note.py'),
                           'musy.note.Note.__gt__': ('note.html#note.__gt__', 'musy/note.py'),
                           'musy.note.Note.__index__': ('note.html#note.__index__', 'musy/note.py'),
                           'musy.note.Note.__init__': ('note.html#note.__init__', 'musy/note.py'),
                           'musy.note.Note.__int__': ('note.html#note.__int__', 'musy/note.py'),
                           'musy.note.Note.__le__': ('note.html#note.__le__', 'musy/note.py'),
                           'musy.note.Note.__lt__': ('note.html#note.__lt__', 'musy/note.py'),
                           'musy.note.Note.__mod__': ('note.html#note.__mod__', 'musy/note.py'),
                           'musy.note.Note.__ne__': ('note.html#note.__ne__', 'musy/note.py'),
                           'musy.note.Note.__repr__': ('note.html#note.__repr__', 'musy/note.py'),
                           'musy.note.Note.__str__': ('note.html#note.__str__', 'musy/note.py'),
                           'musy.note.Note.__sub__': ('note.html#note.__sub__', 'musy/note.py'),
                           'musy.note.Note.from_int': ('note.html#note.from_int', 'musy/note.py'),
                           'musy.note.Note.from_midi': ('note.html#note.from_midi', 'musy/note.py'),
                           'musy.note.Note.get_audio_array': ('note.html#note.get_audio_array', 'musy/note.py'),
                           'musy.note.Note.get_audio_bytes': ('note.html#note.get_audio_bytes', 'musy/note.py'),
                           'musy.note.Note.interval': ('note.html#note.interval', 'musy/note.py'),
                           'musy.note.Note.major': ('note.html#note.major', 'musy/note.py'),
                           'musy.note.Note.midi': ('note.html#note.midi', 'musy/note.py'),
                           'musy.note.Note.minor': ('note.html#note.minor', 'musy/note.py'),
                           'musy.note.Note.play': ('note.html#note.play', 'musy/note.py')},
            'musy.scale': { 'musy.scale.Scale': ('scale.html#scale', 'musy/scale.py'),
                            'musy.scale.Scale.__eq__': ('scale.html#scale.__eq__', 'musy/scale.py'),
                            'musy.scale.Scale.__init__': ('scale.html#scale.__init__', 'musy/scale.py'),
                            'musy.scale.Scale.__iter__': ('scale.html#scale.__iter__', 'musy/scale.py'),
                            'musy.scale.Scale.__len__': ('scale.html#scale.__len__', 'musy/scale.py'),
                            'musy.scale.Scale.__ne__': ('scale.html#scale.__ne__', 'musy/scale.py'),
                            'musy.scale.Scale.__repr__': ('scale.html#scale.__repr__', 'musy/scale.py'),
                            'musy.scale.Scale._shift_abs_semitones': ('scale.html#scale._shift_abs_semitones', 'musy/scale.py'),
                            'musy.scale.Scale._shift_rel_semitones': ('scale.html#scale._shift_rel_semitones', 'musy/scale.py'),
                            'musy.scale.Scale.abs_interval_names': ('scale.html#scale.abs_interval_names', 'musy/scale.py'),
                            'musy.scale.Scale.abs_semitones': ('scale.html#scale.abs_semitones', 'musy/scale.py'),
                            'musy.scale.Scale.available_scales': ('scale.html#scale.available_scales', 'musy/scale.py'),
                            'musy.scale.Scale.from_intervals': ('scale.html#scale.from_intervals', 'musy/scale.py'),
                            'musy.scale.Scale.get_audio_array': ('scale.html#scale.get_audio_array', 'musy/scale.py'),
                            'musy.scale.Scale.get_diatonic_chords': ('scale.html#scale.get_diatonic_chords', 'musy/scale.py'),
                            'musy.scale.Scale.get_modes': ('scale.html#scale.get_modes', 'musy/scale.py'),
                            'musy.scale.Scale.get_notes': ('scale.html#scale.get_notes', 'musy/scale.py'),
                            'musy.scale.Scale.get_scale_names': ('scale.html#scale.get_scale_names', 'musy/scale.py'),
                            'musy.scale.Scale.get_sevenths': ('scale.html#scale.get_sevenths', 'musy/scale.py'),
                            'musy.scale.Scale.get_triads': ('scale.html#scale.get_triads', 'musy/scale.py'),
                            'musy.scale.Scale.play': ('scale.html#scale.play', 'musy/scale.py'),
                            'musy.scale.Scale.play_sevenths': ('scale.html#scale.play_sevenths', 'musy/scale.py'),
                            'musy.scale.Scale.play_triads': ('scale.html#scale.play_triads', 'musy/scale.py'),
                            'musy.scale.Scale.rel_interval_names': ('scale.html#scale.rel_interval_names', 'musy/scale.py'),
                            'musy.scale.Scale.rel_semitones': ('scale.html#scale.rel_semitones', 'musy/scale.py'),
                            'musy.scale.Scale.shift_intvals': ('scale.html#scale.shift_intvals', 'musy/scale.py'),
                            'musy.scale.Scale.to_frame': ('scale.html#scale.to_frame', 'musy/scale.py'),
                            'musy.scale.semi_to_intvals': ('scale.html#semi_to_intvals', 'musy/scale.py')},
            'musy.viz': { 'musy.viz.Guitar': ('viz.html#guitar', 'musy/viz.py'),
                          'musy.viz.Guitar.__ft__': ('viz.html#guitar.__ft__', 'musy/viz.py'),
                          'musy.viz.Piano': ('viz.html#piano', 'musy/viz.py'),
                          'musy.viz.Piano.__call__': ('viz.html#piano.__call__', 'musy/viz.py'),
                          'musy.viz.Piano.__ft__': ('viz.html#piano.__ft__', 'musy/viz.py')}}}
